
/*
 * opengl.c
 *
 * OpenGL routines for Tomb Raider II viewer/editor
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "config.h"
#include "tr2.h"

#ifdef INTELPC
#include <sys/types.h>
/*
 * The #define WIN32_LEAN_AND_MEAN and #include <windows.h> are required for
 * gl.h, glu.h, and glut.h - need to make this more portable...
 */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include "texture.h"
#define OPENGL_FULL                 // tell opengl.h to do all the defines and enums that harness.c doesn't want to see
#define GL_EXTERN                   // tell opengl.h to actually declare its variables
#include "opengl.h"
#include "trmath.h"

/*
 * Macros/Constants
 */
#define ROOM_TRANSPARENCY  0xff  // scale 0 == transparent, 0xFF == opaque
#define EYE_TARGET_OFFSET  100000   // to look down

// Geometric constants
const double Pi      = 3.141592653589;
const double TwoPi   = 3.141592653589 * 2.0;
const double PiOver2 = 3.141592653589 / 2.0;

// Colour constants
const GLfloat White[]         = {  1.0,  1.0,  1.0, 1.0 };
const GLfloat DimWhite[]      = {  0.2,  0.2,  0.2, 1.0 };
const GLfloat Gray[]          = {  0.5,  0.5,  0.5, 1.0 };
const GLfloat Black[]         = {  0.0,  0.0,  0.0, 1.0 };
const GLfloat NegativeBlack[] = { -1.0, -1.0, -1.0, 1.0 };
const GLfloat Red[]           = {  1.0,  0.0,  0.0, 1.0 };
const GLfloat Yellow[]        = {  1.0,  1.0,  0.0, 1.0 };
const GLfloat Green[]         = {  0.0,  1.0,  0.0, 1.0 };
const GLfloat Cyan[]          = {  0.0,  1.0,  1.0, 1.0 };
const GLfloat Blue[]          = {  0.0,  0.0,  1.0, 1.0 };
const GLfloat Magenta[]       = {  1.0,  0.0,  1.0, 1.0 };


/*
 * Data structures
 */

/*
 * Prototypes
 */

/*
 * Data
 */
struct {
   opengl_vector eye;
   opengl_vector target;
   } opengl_camera;
double   tr2_Theta = 0.0;
double   tr2_ViewDistance = 14.0;  // somewhat arbitrary
int      LightingMode = FALSE;
GLfloat  LightingArray[4];
int      WhichLight;
int      CurrentRoom;
int      MeshMode = TRUE;
int      AnimateMeshes = TRUE;
int      DrawStaticMeshes = TRUE;
int      DrawRoomSprites = TRUE;
int      DoPerVertexLighting = TRUE;
int      AnimatedTextureMode = TRUE;
int      BoxesMode = FALSE;
int      PortalMode = FALSE;
int      SingleRoomMode = FALSE;
int      DrawSkyMeshMode = TRUE;
int      WhichSingleRoom = 0;
int      MainViewerWindow = -1;
GLdouble ModelViewMatrix[16],
         ProjectionMatrix[16];
GLint    ViewportArray[4];
int      SkyMesh;
unsigned FrameCycle = 0;
double   EyeTargetOffset = 0;
int      AnimatedTextureTick = 0;
int      ColourArrayEnabled = FALSE;
int      AnimatedTextureCycle = 5;  // MUST be >= 1

unsigned char MessageBuffer[128] = { 0 };

#ifdef INTELPC // actually, this is a Win32 thing...
LARGE_INTEGER PerformanceCounterFrequency = { 1 },
              LastPerformanceCounter = { 0 };
#endif

/*
 * Code
 */
void UpdateTitleBar(void)
{
   char NewTitle[256];
   extern char TRueViewVersion[];

   sprintf(NewTitle, "TRueView %s - %s", TRueViewVersion, opengl_level->FileName);
   glutSetWindowTitle(NewTitle);
}

#ifdef NEVER_COMPILE_THIS  // mouse support is currently disabled
/*
 * mouse click handler
 */
void opengl_mouse_click(int button, int state, int x, int y)
{
   if (button == GLUT_LEFT_BUTTON) {
      if (state == GLUT_DOWN) {
         }
      else { // mouse button up
         }
      }
}

/*
 * mouse motion handler
 */
void opengl_mouse_motion(int x, int y)
{
}
#endif

/*
 * command-line parser and other initialisations
 */
void opengl_init(int *argc, char **argv)
{
   /* create opengl context */
   glutInit(argc, argv);
   /* Default to Fast OpenGL */
   FastOpenGL = TRUE;
   MeshMode = TRUE;
   AnimateMeshes = TRUE;
   DrawStaticMeshes = TRUE;
   DrawRoomSprites = TRUE;
   DoPerVertexLighting = TRUE;
   AnimatedTextureMode = TRUE;
   DrawSkyMeshMode = TRUE;
   AnimatedTextureCycle = 8;     // rather arbitrary means of timing the animated textures... (MUST be >= 1)
   /* see if Slow OpenGL was specified on the command line */
   if (*argc > 2 && argv[1][0] == '-') {
      if (!stricmp(argv[1], "-slow") || !stricmp(argv[1], "-s")) {
         FastOpenGL = FALSE;
         MeshMode = FALSE;
         AnimateMeshes = TRUE;
         DrawStaticMeshes = FALSE;
         DrawRoomSprites = FALSE;
         DoPerVertexLighting = TRUE;   // this doesn't seem to "cost" very much...
         AnimatedTextureMode = FALSE;
         DrawSkyMeshMode = FALSE;
         AnimatedTextureCycle = 2;     // rather arbitrary means of timing the animated textures... (MUST be >= 1)
         }
      argv[1] = argv[2];            // copy the filename into argv[1] (for main())
      }
   printf("Using %s OpenGL settings.\n", (FastOpenGL) ? "Fast" : "Slow");
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

   MainViewerWindow = glutCreateWindow("TRueView TR Level Viewer");

   /* set callback functions */
   glutDisplayFunc(opengl_draw);
   glutReshapeFunc(opengl_resize);
   glutSpecialFunc(opengl_key_special);
   glutKeyboardFunc(opengl_key);
//   glutMouseFunc(opengl_mouse_click);
//   glutMotionFunc(opengl_mouse_motion);

   /* set up Z buffering */
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LESS);

   /* set up culling */
   glEnable(GL_CULL_FACE);
   glFrontFace(GL_CW);

   /* set background to black */
   glClearColor(Black[0], Black[1], Black[2], Black[3]);

   /* disable lighting */
   glDisable(GL_LIGHTING);

   /* set up alpha blending */
   if (FastOpenGL) {
      glEnable(GL_BLEND);
      glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
      glEnable(GL_ALPHA_TEST);
      glAlphaFunc(GL_GREATER, 0);
      }
   else {
      glDisable(GL_BLEND);
      glDisable(GL_ALPHA_TEST);
      }
   glShadeModel(GL_SMOOTH);   // if we use GL_FLAT, the per-vertex lighting looks horrible.

   /*
    * Here we do what we can to speed up the OpenGL rendering process.
    * Some of these are disabled by default;  we disable them here just
    * for completeness (and just in case).  If you want to experiment
    * with any of these, feel free.
    */
   if (FastOpenGL) {
      glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
      glHint(GL_FOG_HINT, GL_NICEST);
      glDisable(GL_COLOR_MATERIAL); // lighting mode only!
      glEnable(GL_DITHER);
      glEnable(GL_POLYGON_SMOOTH);
      }
   else {
      glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
      glHint(GL_FOG_HINT, GL_FASTEST);
      glDisable(GL_COLOR_MATERIAL);
      glDisable(GL_DITHER);
      glDisable(GL_POLYGON_SMOOTH);
      }
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_AUTO_NORMAL);
   glDisable(GL_LOGIC_OP);
   glDisable(GL_TEXTURE_1D);
   glDisable(GL_STENCIL_TEST);
   glDisable(GL_FOG);

   glDisable(GL_NORMALIZE);

   glEnableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_EDGE_FLAG_ARRAY);
   glDisableClientState(GL_COLOR_ARRAY);
   ColourArrayEnabled = FALSE;
   glDisableClientState(GL_NORMAL_ARRAY);

   glPolygonMode(GL_FRONT, GL_FILL);

   /* set image alignment */
//   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
//   glPixelStorei(GL_PACK_ALIGNMENT, 1);

   /* call resize function (to set perspective matrix) */
   glutReshapeWindow(640, 480);

#ifdef INTELPC // actually, this is a Win32 thing...
   // see how many ticks per second the clock generates (for FPS indicator)
   QueryPerformanceFrequency(&PerformanceCounterFrequency);
#endif

   /* ask for redraw */
   glutPostRedisplay();
}

/*
 * opengl_print() displays status information in the upper-left corner of the window.
 * It's crude, but effective.
 */
void opengl_print(unsigned char *string)
{
   void *font = GLUT_BITMAP_HELVETICA_18;
   int i;
   char buf[128],
        buf2[64];
#ifdef INTELPC // actually, this is a Win32 thing...
   LARGE_INTEGER Perf;
   double fps;
#endif

   glDisable(GL_TEXTURE_2D);
   glDisable(GL_LIGHTING);
   glDisableClientState(GL_COLOR_ARRAY);
   ColourArrayEnabled = FALSE;
   glDisableClientState(GL_TEXTURE_COORD_ARRAY);
   glMatrixMode(GL_PROJECTION);

#define RASTER_MULT 2.8 // trial-and-error value

   // display whatever string was passed in
   glPushMatrix();
   glTranslated(-2.41 * RASTER_MULT, -1.75 * RASTER_MULT, 0);
   if (EyeTargetOffset != 0) {
      glTranslated(-48200, -35000, 0);
      }
   glRasterPos3f(opengl_camera.target.x, opengl_camera.target.y, opengl_camera.target.z);
   glColor3fv(White);
   for (i = 0; string[i]; ++i) {
      glutBitmapCharacter(font, string[i]);
      }
   glPopMatrix();

   // display the position and orientation of the camera
   font = GLUT_BITMAP_HELVETICA_12;
   glPushMatrix();
   glTranslated(-1.0 * RASTER_MULT, -1.75 * RASTER_MULT, 0);
   sprintf(buf, "   X: %3.0f  Y: %3.0f  Z: %3.0f  Rot: %3.0f", opengl_camera.eye.x, opengl_camera.eye.y, opengl_camera.eye.z, tr2_Theta * (180.0 / Pi));

   // if we're only displaying one room, tell which one it is
   if (SingleRoomMode) {
      sprintf(buf2, "  Displaying Room %d", WhichSingleRoom);
      strcat(buf, buf2);
      }
   
   // update alternate room indicator
   i = WhichRoomCamera();
   if (i >= 0) {
      sprintf(buf2, "  AltRoom %d", opengl_level->Rooms[i].AlternateRoom);
      strcat(buf, buf2);
      }

#ifdef INTELPC // actually, this is a Win32 thing...
   // update FPS display
   QueryPerformanceCounter(&Perf);
   fps = Perf.QuadPart - LastPerformanceCounter.QuadPart;   // ticks since last update
   LastPerformanceCounter = Perf;
   fps = (double)PerformanceCounterFrequency.QuadPart / fps;   // ticks/sec / ticks = 1/sec
   sprintf(buf2, "  FPS %4.1f", fps);
   strcat(buf, buf2);
#endif

   // write the string
   for (i = 0; buf[i]; ++i) {
      glutBitmapCharacter(font, buf[i]);
      }
   glPopMatrix();
}

/*
 * As its name implies, ResetEyePosition() resets the camera to the initial position
 * for the level.
 */
void ResetEyePosition(void)
{
   int i;

   /* set camera position */
   for (i = 0; i < opengl_level->NumItems; ++i) {
      if (opengl_level->Items[i].ObjectID == 0)
         break;
      }
   if (i == opengl_level->NumItems)
      i = 0;
   opengl_camera.eye.x = opengl_level->Items[i].x;
   opengl_camera.eye.y = opengl_level->Items[i].y - 1000;
   opengl_camera.eye.z = opengl_level->Items[i].z;
   tr2_Theta = ((opengl_level->Items[i].Angle >> 14) & 0x03) * PiOver2;
}

/*
 * As its name implies, DisplayEyePosition displays the position of the camera.
 * (If not using stdout, it only displays the current room number)
 */
void DisplayEyePosition(void)
{
   int which;

#ifdef DO_STDOUT
   printf("Eye at (%f, %f, %f) theta %f\n",
          opengl_camera.eye.x, opengl_camera.eye.y, opengl_camera.eye.z,
          tr2_Theta);
#endif
   which = WhichRoomCamera();
   if (which < 0)
      strcpy((char *)MessageBuffer, "Outside all rooms");
   else {
      bitu32 sector;

      sector = ((((int)opengl_camera.eye.x - opengl_level->Rooms[which].info.x) / 1024) * opengl_level->Rooms[which].NumZsectors) +
               (((int)opengl_camera.eye.z - opengl_level->Rooms[which].info.z) / 1024);
      sprintf((char *)MessageBuffer, "Room %d, Sector %d", which, sector);
      }
}

void UpdateMenuSelections(void)
{
   char buffer[80];
   int MenuItem;

#define RIGHT_CLICK_MENU_CHECK_MARK '*'

   MenuItem = 1;
   sprintf(buffer, ".%c Enable Per-Vertex Lighting [V]", (DoPerVertexLighting) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_VERTEX_LIGHTING);

   sprintf(buffer, ".%c Display Item Meshes [M]", (MeshMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_ITEM_MESHES);
   
   sprintf(buffer, ".%c Animate Meshes [W]", (AnimateMeshes) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_ANIMATED_MESHES);

   sprintf(buffer, ".%c Display Static Meshes [S]", (DrawStaticMeshes) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_STATIC_MESHES);

   sprintf(buffer, ".%c Display Room Sprites [Q]", (DrawRoomSprites) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_ROOM_SPRITES);

   if (opengl_level->NumAnimatedTextures != 0) {
      sprintf(buffer, ".%c Animate Textures [Y]", (AnimatedTextureMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
      glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_ANIMATED_TEXTURES);
      }
   else {
      glutChangeToMenuEntry(MenuItem++, ".", GLUT_MENU_TOGGLE_ANIMATED_TEXTURES);
      }
   
   sprintf(buffer, ".%c Enable Sky Mesh Display [O]", (DrawSkyMeshMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_SKYMESH_DISPLAY);

   sprintf(buffer, ".%c Enable Visible Portals [P]", (PortalMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_PORTALS);
   
   if (SingleRoomMode) {
      sprintf(buffer, ".%c Single Room Display [R]", RIGHT_CLICK_MENU_CHECK_MARK);
      glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_SINGLE_ROOM_DISPLAY);
      glutChangeToMenuEntry(MenuItem++, ".  Display Next Room [+]", GLUT_MENU_NEXT_SINGLE_ROOM);
      glutChangeToMenuEntry(MenuItem++, ".  Display Previous Room [-]", GLUT_MENU_PREVIOUS_SINGLE_ROOM);
      }
   else {
      sprintf(buffer, ".  Single Room Display [R]");
      glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_SINGLE_ROOM_DISPLAY);
      glutChangeToMenuEntry(MenuItem++, ".", 0);
      glutChangeToMenuEntry(MenuItem++, ".", 0);
      }
   
   sprintf(buffer, ".%c Enable Boxes[] Display [B]", (BoxesMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_BOXES_DISPLAY);

   // 12 = save level
   MenuItem++;

   // 13 = dump level
   MenuItem++;

   sprintf(buffer, ".%c Enable Alpha Transparency [A]", glIsEnabled(GL_ALPHA_TEST) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_ALPHA);

   sprintf(buffer, ".%c Enable Fog [F]", glIsEnabled(GL_FOG) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_FOG);

   sprintf(buffer, ".%c Enable Face Culling [C]", glIsEnabled(GL_CULL_FACE) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_FACE_CULLING);
   
   sprintf(buffer, ".%c Enable Lighting [L]", (LightingMode) ? RIGHT_CLICK_MENU_CHECK_MARK : ' ');
   glutChangeToMenuEntry(MenuItem++, buffer, GLUT_MENU_TOGGLE_LIGHTING);
   
   // 18 = reset window size
//   MenuItem++;

   // 19 = reset camera position
//   MenuItem++;
}

/*
 * Right-click GLUT menu handler.  <value> indicates what the user selected.
 */
void Menu_Base(int value)
{
   switch (value) {
      case GLUT_MENU_TOGGLE_ANIMATED_MESHES :
         AnimateMeshes = 1 - AnimateMeshes;
         FrameCycle = 0;
         if (AnimateMeshes) {
            strcpy((char *)MessageBuffer, "Animated Meshes ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Animated Meshes OFF");
            }
         break;
      case GLUT_MENU_TOGGLE_SINGLE_ROOM_DISPLAY :
         SingleRoomMode = 1 - SingleRoomMode;
         strcpy((char *)MessageBuffer, (SingleRoomMode) ? "Single Room Mode" : "All Rooms Mode");
         break;
      case GLUT_MENU_NEXT_SINGLE_ROOM :
         if (SingleRoomMode) {
            if (++WhichSingleRoom >= opengl_level->NumRooms)
               WhichSingleRoom = 0;
            }
         break;
      case GLUT_MENU_PREVIOUS_SINGLE_ROOM :
         if (SingleRoomMode) {
            if (--WhichSingleRoom < 0)
               WhichSingleRoom = opengl_level->NumRooms - 1;
            }
         break;
      case GLUT_MENU_TOGGLE_ANIMATED_TEXTURES :
         if (opengl_level->EngineVersion != TombRaider_4 && opengl_level->NumAnimatedTextures != 0) {
            AnimatedTextureMode = 1 - AnimatedTextureMode;
            strcpy((char *)MessageBuffer, (AnimatedTextureMode) ? "Animated Textures ON" : "Animated Textures OFF");
            }
         break;
      case GLUT_MENU_TOGGLE_BOXES_DISPLAY :
         BoxesMode = 1 - BoxesMode;
         strcpy((char *)MessageBuffer, (BoxesMode) ? "Boxes[] ON" : "Boxes[] OFF");
         break;
      case GLUT_MENU_TOGGLE_SKYMESH_DISPLAY :
         DrawSkyMeshMode = 1 - DrawSkyMeshMode;
         strcpy((char *)MessageBuffer, (DrawSkyMeshMode) ? "SkyMesh ON" : "SkyMesh OFF");
         break;
      case GLUT_MENU_SAVE_LEVEL :
         WriteTR2level(opengl_level, "TestLvl.TR2");
         strcpy((char *)MessageBuffer, "Level saved as TestLvl.TR2");
         break;
      case GLUT_MENU_DUMP_LEVEL :
         DumpTR2level(opengl_level, "TestLvl.out");
         strcpy((char *)MessageBuffer, "Level dumped as TestLvl.out");
         break;
      case GLUT_MENU_TOGGLE_STATIC_MESHES :
         DrawStaticMeshes = 1 - DrawStaticMeshes;
         strcpy((char *)MessageBuffer, (DrawStaticMeshes) ? "StaticMeshes ON" : "StaticMeshes OFF");
         break;
      case GLUT_MENU_TOGGLE_ROOM_SPRITES :
         DrawRoomSprites = 1 - DrawRoomSprites;
         strcpy((char *)MessageBuffer, (DrawRoomSprites) ? "RoomSprites ON" : "RoomSprites OFF");
         break;
      case GLUT_MENU_TOGGLE_VERTEX_LIGHTING :
         DoPerVertexLighting = 1 - DoPerVertexLighting;
         strcpy((char *)MessageBuffer, (DoPerVertexLighting) ? "Per-Vertex Lighting ON" : "Per-Vertex Lighting OFF");
         glShadeModel((DoPerVertexLighting) ? GL_SMOOTH : GL_FLAT);
         break;
      case GLUT_MENU_RESET_WINDOW_SIZE :
         opengl_resize(640, 480);
         glutReshapeWindow(640, 480);
         break;
      case GLUT_MENU_TOGGLE_ALPHA :
         if (glIsEnabled(GL_ALPHA_TEST)) {
            glDisable(GL_ALPHA_TEST);
            glDisable(GL_BLEND);
            strcpy((char *)MessageBuffer, "Alpha Transparency OFF");
            }
         else {
            glEnable(GL_ALPHA_TEST);
            glAlphaFunc(GL_GREATER, 0);
            glEnable(GL_BLEND);
            glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
            strcpy((char *)MessageBuffer, "Alpha Transparency ON");
            }
         break;
      case GLUT_MENU_TOGGLE_FOG :
         if (glIsEnabled(GL_FOG)) {
            strcpy((char *)MessageBuffer, "Fog OFF");
            glDisable(GL_FOG);
            }
         else {
            strcpy((char *)MessageBuffer, "Fog ON");
            glEnable(GL_FOG);
            glFogf(GL_FOG_MODE, GL_EXP2);
            glFogf(GL_FOG_DENSITY, 0.00008);
            glFogf(GL_FOG_START, 30000.0);
            glFogf(GL_FOG_END, 50000.0);
            glFogfv(GL_FOG_COLOR, Black);
            }
         break;
      case GLUT_MENU_TOGGLE_ITEM_MESHES :
         MeshMode = 1 - MeshMode;
         strcpy((char *)MessageBuffer, (MeshMode) ? "Meshes ON" : "Meshes OFF");
         break;
      case GLUT_MENU_TOGGLE_FACE_CULLING :
         if (glIsEnabled(GL_CULL_FACE)) {
            glDisable(GL_CULL_FACE);
            strcpy((char *)MessageBuffer, "Culling OFF");
            }
         else {
            glEnable(GL_CULL_FACE);
            strcpy((char *)MessageBuffer, "Culling ON");
            }
         break;
      case GLUT_MENU_TOGGLE_LIGHTING :
         LightingMode = 1 - LightingMode;
         if (LightingMode) {
            strcpy((char *)MessageBuffer, "Lighting ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Lighting OFF");
            glDisable(GL_LIGHTING);
            }
         break;
      case GLUT_MENU_TOGGLE_PORTALS :
         PortalMode = 1 - PortalMode;
         if (PortalMode) {
            strcpy((char *)MessageBuffer, "Visible Portals ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Visible Portals OFF");
            }
         break;
      case GLUT_MENU_RESET_CAMERA :
         ResetEyePosition();
         break;
      case GLUT_MENU_EXIT :
         exit(0);
         break;
      }
   UpdateMenuSelections(); // make the next menu display correct
}

/*
 * OpenGL entrypoint.
 */
void opengl_main(TR2_Level *level)
{
   /* simplify our lives - make level global */
   opengl_level = level;

   SkyMesh = -1;
   if (opengl_level->EngineVersion == TombRaider_2) {
      int i;
      for (i = 0; i < (int)opengl_level->NumMoveables; ++i) {
         if (opengl_level->Moveables[i].ObjectID == 254) {
            SkyMesh = opengl_level->Moveables[i].StartingMesh;
            break;
            }
         }
      }
   else if (opengl_level->EngineVersion == TombRaider_3) {
      int i;
      for (i = 0; i < (int)opengl_level->NumMoveables; ++i) {
         if (opengl_level->Moveables[i].ObjectID == 355) {
            SkyMesh = opengl_level->Moveables[i].StartingMesh;
            break;
            }
         }
      }
   else if (opengl_level->EngineVersion == TombRaider_4) {
      AnimatedTextureMode = FALSE;  // this doesn't work (yet) under TR4, so don't bother
      }

   /* set up texture table */
   OpenGL_TextureListOffset = 0;
   OpenGL_TextureList = NULL;
   OpenGL_TextureList = (opengl_texture *)TR2alloc(sizeof(opengl_texture) * (opengl_level->NumObjectTextures + opengl_level->NumSpriteTextures));

   /* default mipmapping to ON */
   if (FastOpenGL) {
      opengl_mipmap_textures = TRUE;
      }
   else {
      /* default mipmapping to OFF */
      opengl_mipmap_textures = FALSE;  // turning this ON makes things look better, but it's SLOW...
      }

   ResetEyePosition();
   opengl_camera.target.x = tr2_ViewDistance;
   opengl_camera.target.y = 0.0;
   opengl_camera.target.z = 0.0;

   /* set up OpenGL meshes */
   opengl_create_all_meshes();
   opengl_init_sprite_textures();

   // set up menus (just use dummy names; UpdateMenuSelections() will fill in the "real" ones)
   glutCreateMenu(Menu_Base);

   glutAddMenuEntry(".  Per-Vertex", GLUT_MENU_TOGGLE_VERTEX_LIGHTING);
   glutAddMenuEntry(".  Items", GLUT_MENU_TOGGLE_ITEM_MESHES);
   glutAddMenuEntry(".  AnimMeshes", GLUT_MENU_TOGGLE_ANIMATED_MESHES);
   glutAddMenuEntry(".  Static Mesh", GLUT_MENU_TOGGLE_STATIC_MESHES);
   glutAddMenuEntry(".  RoomSprite", GLUT_MENU_TOGGLE_ROOM_SPRITES);
   glutAddMenuEntry(".  AnimTextures", GLUT_MENU_TOGGLE_ANIMATED_TEXTURES);
   glutAddMenuEntry(".  SkyMesh", GLUT_MENU_TOGGLE_SKYMESH_DISPLAY);
   glutAddMenuEntry(".  Portals", GLUT_MENU_TOGGLE_PORTALS);
   glutAddMenuEntry(".  Single Room", GLUT_MENU_TOGGLE_SINGLE_ROOM_DISPLAY);
   glutAddMenuEntry(".  Next Room", GLUT_MENU_NEXT_SINGLE_ROOM);
   glutAddMenuEntry(".  Previous Room", GLUT_MENU_PREVIOUS_SINGLE_ROOM);
   glutAddMenuEntry(".  Boxes", GLUT_MENU_TOGGLE_BOXES_DISPLAY);
   glutAddMenuEntry(".  Save Level to TESTLVL.TR2 [F5]", GLUT_MENU_SAVE_LEVEL);
   glutAddMenuEntry(".  Dump Level to TESTLVL.OUT [$]", GLUT_MENU_DUMP_LEVEL);
   glutAddMenuEntry(".  Alpha", GLUT_MENU_TOGGLE_ALPHA);
   glutAddMenuEntry(".  Fog", GLUT_MENU_TOGGLE_FOG);
   glutAddMenuEntry(".  Culling", GLUT_MENU_TOGGLE_FACE_CULLING);
   glutAddMenuEntry(".  Lighting", GLUT_MENU_TOGGLE_LIGHTING);
   glutAddMenuEntry(".  Reset Window Size [`]", GLUT_MENU_RESET_WINDOW_SIZE);
   glutAddMenuEntry(".  Reset Camera to Initial Position [.]", GLUT_MENU_RESET_CAMERA);
   glutAddMenuEntry(".  ---------------", 0);
   glutAddMenuEntry(".  Exit TR2 Viewer [ESC]", GLUT_MENU_EXIT);
   glutAttachMenu(GLUT_RIGHT_BUTTON);

   UpdateMenuSelections();

   UpdateTitleBar();
   DisplayEyePosition();

   /* go into glut main loop */
   glutMainLoop();
}

/*
 * GLUT non-ASCII key handler (e.g. F-keys, arrow keys, etc.)
 */
void opengl_key_special(int key, int x, int y)
{
   switch (key) {
      case GLUT_KEY_END :   // flip camera 180 degrees (simulate the TR2 "ROLL" key)
         tr2_Theta += Pi;
         if (tr2_Theta > Pi)
            tr2_Theta -= TwoPi;
         opengl_camera.target.z = (tr2_ViewDistance * cos(tr2_Theta)) + opengl_camera.eye.z;
         opengl_camera.target.x = (tr2_ViewDistance * sin(tr2_Theta)) + opengl_camera.eye.x;
         DisplayEyePosition();
         break;

      case GLUT_KEY_F12 :  // exit the viewer
         exit(0);
         break;

      case GLUT_KEY_F5 :  // save level
         key = WriteTR2level(opengl_level, "TestLvl.TR2");
         strcpy((char *)MessageBuffer, "Level saved as TestLvl.TR2");
         break;

      case GLUT_KEY_DOWN:  // move camera backward
      	opengl_camera.eye.z -= (TRANSLATE_DELTA * cos(tr2_Theta));
   	   opengl_camera.eye.x -= (TRANSLATE_DELTA * sin(tr2_Theta));
         DisplayEyePosition();
         break;

      case GLUT_KEY_UP:    // move camera forward
      	opengl_camera.eye.z += (TRANSLATE_DELTA * cos(tr2_Theta));
   	   opengl_camera.eye.x += (TRANSLATE_DELTA * sin(tr2_Theta));
         DisplayEyePosition();
         break;

      case GLUT_KEY_RIGHT: // rotate camera right
         tr2_Theta += ROTATE_DELTA;
         if (tr2_Theta > Pi)
            tr2_Theta -= TwoPi;
         opengl_camera.target.z = (tr2_ViewDistance * cos(tr2_Theta)) + opengl_camera.eye.z;
         opengl_camera.target.x = (tr2_ViewDistance * sin(tr2_Theta)) + opengl_camera.eye.x;
         DisplayEyePosition();
         break;

      case GLUT_KEY_LEFT:  // rotate camera left
         tr2_Theta -= ROTATE_DELTA;
         if (tr2_Theta < -Pi)
            tr2_Theta += TwoPi;
      	opengl_camera.target.z = (tr2_ViewDistance * cos(tr2_Theta)) + opengl_camera.eye.z;
      	opengl_camera.target.x = (tr2_ViewDistance * sin(tr2_Theta)) + opengl_camera.eye.x;
         DisplayEyePosition();
         break;

      case GLUT_KEY_PAGE_UP:     // move camera up
         opengl_camera.eye.y -= TRANSLATE_DELTA / 2.0;   // use smaller steps on vertical changes
         opengl_camera.target.y -= TRANSLATE_DELTA / 2.0;
         DisplayEyePosition();
         break;

      case GLUT_KEY_PAGE_DOWN:   // move camera down
         opengl_camera.eye.y += TRANSLATE_DELTA / 2.0;
         opengl_camera.target.y += TRANSLATE_DELTA / 2.0;
         DisplayEyePosition();
         break;
      }
}

/*
 * GLUT ASCII key handler
 */
void opengl_key(unsigned char key, int x, int y)
{
   int key_upper;

   /* convert key to upper case */
   key_upper = toupper((int)key); 

   switch (key_upper) {
      case 'W' :  // toggle animated meshes
         AnimateMeshes = 1 - AnimateMeshes;
         FrameCycle = 0;
         if (AnimateMeshes) {
            strcpy((char *)MessageBuffer, "Animated Meshes ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Animated Meshes OFF");
            }
         break;
      case 'R' :  // toggle single room display
         SingleRoomMode = 1 - SingleRoomMode;
         strcpy((char *)MessageBuffer, (SingleRoomMode) ? "Single Room Mode" : "All Rooms Mode");
         break;
      case '+' :
         if (SingleRoomMode) {
            if (++WhichSingleRoom >= opengl_level->NumRooms)
               WhichSingleRoom = 0;
            }
         break;
      case '-' :
         if (SingleRoomMode) {
            if (--WhichSingleRoom < 0)
               WhichSingleRoom = opengl_level->NumRooms - 1;
            }
         break;
      case 'Y' :  // toggle animated textures
         if (opengl_level->EngineVersion != TombRaider_4 && opengl_level->NumAnimatedTextures != 0) {
            AnimatedTextureMode = 1 - AnimatedTextureMode;
            strcpy((char *)MessageBuffer, (AnimatedTextureMode) ? "Animated Textures ON" : "Animated Textures OFF");
            }
         break;
      case '%' :  // look down (or not)
         if (EyeTargetOffset == 0)
            EyeTargetOffset = EYE_TARGET_OFFSET;
         else
            EyeTargetOffset = 0;
         break;
      case 'B' :  // draw Boxes[]
         BoxesMode = 1 - BoxesMode;
         strcpy((char *)MessageBuffer, (BoxesMode) ? "Boxes[] ON" : "Boxes[] OFF");
         break;
      case '$' :  // save level.out
         key_upper = DumpTR2level(opengl_level, "level.out");
         break;
      case 'M' : // toggle MESH mode
         MeshMode = 1 - MeshMode;
         strcpy((char *)MessageBuffer, (MeshMode) ? "Meshes ON" : "Meshes OFF");
         break;
      case 'O' : // toggle SKYMESH mode
         DrawSkyMeshMode = 1 - DrawSkyMeshMode;
         strcpy((char *)MessageBuffer, (DrawSkyMeshMode) ? "SkyMesh ON" : "SkyMesh OFF");
         break;
      case 'S' : // toggle STATIC MESH mode
         DrawStaticMeshes = 1 - DrawStaticMeshes;
         strcpy((char *)MessageBuffer, (DrawStaticMeshes) ? "StaticMeshes ON" : "StaticMeshes OFF");
         break;
      case 'Q' : // toggle ROOM SPRITE mode
         DrawRoomSprites = 1 - DrawRoomSprites;
         strcpy((char *)MessageBuffer, (DrawRoomSprites) ? "RoomSprites ON" : "RoomSprites OFF");
         break;
      case 'V' : // toggle per-vertex lighting mode
         DoPerVertexLighting = 1 - DoPerVertexLighting;
         strcpy((char *)MessageBuffer, (DoPerVertexLighting) ? "Per-Vertex Lighting ON" : "Per-Vertex Lighting OFF");
         glShadeModel((DoPerVertexLighting) ? GL_SMOOTH : GL_FLAT);
         break;
      case 'F' : // toggle fog mode
         if (glIsEnabled(GL_FOG)) {
            strcpy((char *)MessageBuffer, "Fog OFF");
            glDisable(GL_FOG);
            }
         else {
            strcpy((char *)MessageBuffer, "Fog ON");
            glEnable(GL_FOG);
            glFogf(GL_FOG_MODE, GL_EXP2);
            glFogf(GL_FOG_DENSITY, 0.00008);
            glFogf(GL_FOG_START, 30000.0);
            glFogf(GL_FOG_END, 50000.0);
            glFogfv(GL_FOG_COLOR, Black);
            }
         break;
      case 'L' : // toggle lighting mode
         LightingMode = 1 - LightingMode;
         if (LightingMode) {
            strcpy((char *)MessageBuffer, "Lighting ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Lighting OFF");
            glDisable(GL_LIGHTING);
            }
         break;
      case 'A' : // toggle alpha testing
         if (glIsEnabled(GL_ALPHA_TEST)) {
            glDisable(GL_ALPHA_TEST);
            glDisable(GL_BLEND);
            strcpy((char *)MessageBuffer, "Alpha Transparency OFF");
            }
         else {
            glEnable(GL_ALPHA_TEST);
            glAlphaFunc(GL_GREATER, 0);
            glEnable(GL_BLEND);
            glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
            strcpy((char *)MessageBuffer, "Alpha Transparency ON");
            }
         break;
      case 'C' : // toggle face culling
         if (glIsEnabled(GL_CULL_FACE)) {
            glDisable(GL_CULL_FACE);
            strcpy((char *)MessageBuffer, "Culling OFF");
            }
         else {
            glEnable(GL_CULL_FACE);
            strcpy((char *)MessageBuffer, "Culling ON");
            }
         break;
      case 'P' : // toggle portal mode
         PortalMode = 1 - PortalMode;
         if (PortalMode) {
            strcpy((char *)MessageBuffer, "Visible Portals ON");
            }
         else {
            strcpy((char *)MessageBuffer, "Visible Portals OFF");
            }
         break;
      case ESCAPE :
         exit(0);
         break;
      case '.' : // reset to starting position/angle
         ResetEyePosition();
         break;
      case '`' : // resize to 640x480 (in case you accidentally resized the window and it looks awful)
         opengl_resize(640, 480);
         glutReshapeWindow(640, 480);
         break;
   }
   UpdateMenuSelections(); // keep menu display in sync with changes made by keyboard
}

/*
 * OpenGL callback function - handle window resizing
 */
void opengl_resize(int width, int height)
{
   /* change perspective matrix and viewport */
   glViewport(0, 0, width, height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   // adjust clipping
   gluPerspective(FOV_Y, ((GLdouble)width) / ((GLdouble)height), CLIP_NEAR * 100, CLIP_FAR * 100);

   /* redraw at new size */
   glGetDoublev(GL_PROJECTION_MATRIX, ProjectionMatrix);
   glGetIntegerv(GL_VIEWPORT, ViewportArray);
   glutPostRedisplay();
}

/*
 * Scan the bounding boxes in the room mesh list to see where camera.eye is now.
 * Note that we return the first match, irrespective of any AlternateRoom...
 */
int WhichRoom(GLfloat x, GLfloat y, GLfloat z)  // figure out where (x, y, z) is
{
   int i;

   for (i = 0; i < opengl_level->NumRooms; ++i) {
      if (x >= opengl_item.RoomMeshes[i]->BoundingBoxLow.x &&
          x <= opengl_item.RoomMeshes[i]->BoundingBoxHigh.x &&
          y >= opengl_item.RoomMeshes[i]->BoundingBoxLow.y &&
          y <= opengl_item.RoomMeshes[i]->BoundingBoxHigh.y &&
          z >= opengl_item.RoomMeshes[i]->BoundingBoxLow.z &&
          z <= opengl_item.RoomMeshes[i]->BoundingBoxHigh.z)
         return(i);
      }
   return(-1); // unknown room
}

int WhichRoomCamera(void)  // figure out where camera.eye is
{
   return(WhichRoom(opengl_camera.eye.x, opengl_camera.eye.y, opengl_camera.eye.z));
}

/*
 * Scan adjacent room list for current room to see if <which> is adjacent.
 */
int IsAdjacent(int which)
{
   int i;

   if (CurrentRoom < 0)
      return(FALSE);
   for (i = 0; i < opengl_item.RoomMeshes[CurrentRoom]->NumAdjacentRooms; ++i) {
      if (which == opengl_item.RoomMeshes[CurrentRoom]->AdjacentRooms[i])
         return(TRUE);
      }
   return(FALSE);
}

/*
 * Set up point lighting for current room (and as many adjacent rooms as possible)
 *
 *** THIS IS BADLY BROKEN.  IT IS CURRENTLY DISABLED.
 */
void Setup_Lighting(void)
{
   int i,
       current,
       zz,
       xx;
   float intensity = 0.0,
         diffuse = 0.0,
         temp;

return;  /////////// DON'T EVEN BOTHER (for now)

   if (LightingMode) {
      if ((current = CurrentRoom) < 0)
         current = 0;   ///////////// KLUDGE for when we wander out of range
      glEnable(GL_LIGHTING);
      glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 0);
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, White);
      glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, White);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, DimWhite); // Black
      WhichLight = 0;
      for (zz = 0; zz < opengl_item.RoomMeshes[current]->NumAdjacentRooms; ++zz) {
         xx = opengl_item.RoomMeshes[current]->AdjacentRooms[zz];
         for (i = 0; i < opengl_level->Rooms[xx].NumLights && WhichLight < 8; ++i) {
            glEnable(GL_LIGHT0 + WhichLight);
            glLightf(GL_LIGHT0 + WhichLight, GL_CONSTANT_ATTENUATION, 1.0 /* 0.0 */);
            temp = opengl_level->Rooms[xx].Lights[i].Fade1;
            temp /= 10000.0; // 144000.0;
            glLightf(GL_LIGHT0 + WhichLight, GL_QUADRATIC_ATTENUATION /* GL_LINEAR_ATTENUATION */, temp /* 0.01 */);
            // intensity of light
            LightingArray[0] = (opengl_level->Rooms[xx].Lights[i].Intensity1 / 409.60); // r
            LightingArray[1] = LightingArray[0]; // g
            LightingArray[2] = LightingArray[0]; // b
            if (opengl_level->EngineVersion == TombRaider_3) {
               LightingArray[0] = opengl_level->Rooms[xx].RoomLightColour.Red; // r
               LightingArray[1] = opengl_level->Rooms[xx].RoomLightColour.Green; // g
               LightingArray[2] = opengl_level->Rooms[xx].RoomLightColour.Blue; // b
               }
            LightingArray[3] = 1.0; // alpha
            intensity = LightingArray[0]; // for printf() that follows
            glLightfv(GL_LIGHT0 + WhichLight, GL_AMBIENT /* GL_DIFFUSE */, LightingArray);
            // position of light
            LightingArray[0] = opengl_level->Rooms[xx].Lights[i].x; // x
            LightingArray[1] = opengl_level->Rooms[xx].Lights[i].y; // y
            LightingArray[2] = opengl_level->Rooms[xx].Lights[i].z; // z
            LightingArray[3] = 1.0; // clipping plane
            glLightfv(GL_LIGHT0 + WhichLight, GL_POSITION, LightingArray);
            ++WhichLight;
            }
         }
      }
//   else {
//      glDisable(GL_LIGHTING);
//      }
}

int IsBehind(int x, int z) // determines if (x, z) is "behind" camera eye
{
   double bTheta,
          bCameraX,
          bCameraZ,
          Distance;

   // set up a "virtual camera" a huge distance behind us
   bTheta = tr2_Theta + Pi;
   if (bTheta > Pi)
      bTheta -= TwoPi;
   bCameraX = (65536.0 * sin(bTheta)) + opengl_camera.eye.x;   // 64k is a fair distance away...
   bCameraZ = (65536.0 * cos(bTheta)) + opengl_camera.eye.z;

   bCameraX -= x;
   bCameraZ -= z;
   Distance = sqrt((bCameraX * bCameraX) + (bCameraZ * bCameraZ));
   return(Distance < 65536.0);
}

double vCameraX, vCameraY, vCameraZ, vTheta;

int RoomDepthCompare(const int *rmPtr1, const int *rmPtr2) // this gets called a LOT - needs optimisation
{
   double dx,
          dy,
          dz,
          depth1,
          dtmp,
          depth2,
          tempval;
   int    RoomMesh1 = *rmPtr1,
          RoomMesh2 = *rmPtr2;

   // calculate distance to room 1 (farthest point in all eight corners of room)
   dx = vCameraX - opengl_level->Rooms[RoomMesh1].info.x;
   dx *= dx;
   tempval = dx;
   dy = vCameraY - opengl_level->Rooms[RoomMesh1].info.yBottom;
   dy *= dy;
   dz = vCameraZ - opengl_level->Rooms[RoomMesh1].info.z;
   dz *= dz;
   depth1 = sqrt(dx + dy + dz);  // corner 1

   dx = vCameraX - (opengl_level->Rooms[RoomMesh1].info.x + (1024 * opengl_level->Rooms[RoomMesh1].NumXsectors));
   dx *= dx;
   dtmp = sqrt(dx + dy + dz);  // corner 2
   if (dtmp > depth1)
      depth1 = dtmp;

   dz = vCameraZ - (opengl_level->Rooms[RoomMesh1].info.z + (1024 * opengl_level->Rooms[RoomMesh1].NumZsectors));
   dz *= dz;
   dtmp = sqrt(dx + dy + dz);  // corner 4
   if (dtmp > depth1)
      depth1 = dtmp;

   dx = tempval;
   dtmp = sqrt(dx + dy + dz);  // corner 3
   if (dtmp > depth1)
      depth1 = dtmp;

   // calculate distance to room 2 (farthest point in all eight corners of room)
   dx = vCameraX - opengl_level->Rooms[RoomMesh2].info.x;
   dx *= dx;
   tempval = dx;
   dy = vCameraY - opengl_level->Rooms[RoomMesh2].info.yBottom;
   dy *= dy;
   dz = vCameraZ - opengl_level->Rooms[RoomMesh2].info.z;
   dz *= dz;
   depth2 = sqrt(dx + dy + dz);

   dx = vCameraX - (opengl_level->Rooms[RoomMesh2].info.x + (1024 * opengl_level->Rooms[RoomMesh2].NumXsectors));
   dx *= dx;
   dtmp = sqrt(dx + dy + dz);  // corner 2
   if (dtmp > depth2)
      depth2 = dtmp;

   dz = vCameraZ - (opengl_level->Rooms[RoomMesh2].info.z + (1024 * opengl_level->Rooms[RoomMesh2].NumZsectors));
   dz *= dz;
   dtmp = sqrt(dx + dy + dz);  // corner 4
   if (dtmp > depth2)
      depth2 = dtmp;

   dx = tempval;
   dtmp = sqrt(dx + dy + dz);  // corner 3
   if (dtmp > depth2)
      depth2 = dtmp;

   if (depth1 > depth2)
      return(1);
   else if (depth1 < depth2)
      return(-1);
   return(0);
   // return the relative distance between them
//   return((int)(depth1 - depth2));
}

void opengl_draw_sprite(double x, double y, double z, int texture, double scale)
{
   // make a fake mesh & rotate it orthogonal to camera
   opengl_vector SpriteMesh[4];
   opengl_texture *tex;
   opengl_texture_coord SpriteTex[4];
   double Width,
          Height;

   if (IsBehind((int)x, (int)z))  // don't bother drawing things that are behind the camera
      return;
   tex = opengl_item.SpriteTextures[texture];
   Width = tex->w * scale;   // scale it up a bit (sprites are rather tiny)
   Height = tex->h * scale;
   SpriteMesh[0].x = -Width / 2.0;
   SpriteMesh[1].x = -Width / 2.0;
   SpriteMesh[2].x = Width / 2.0;
   SpriteMesh[3].x = Width / 2.0;
   SpriteMesh[0].y = 0;
   SpriteMesh[1].y = -Height;
   SpriteMesh[2].y = -Height;
   SpriteMesh[3].y = 0;;
   SpriteMesh[0].z = 0;
   SpriteMesh[1].z = 0;
   SpriteMesh[2].z = 0;
   SpriteMesh[3].z = 0;
   SpriteTex[3].s = (double)(tex->tex->x + tex->tex->Width) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[3].t = (double)(tex->tex->y + tex->tex->Height) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[2].s = (double)(tex->tex->x + tex->tex->Width) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[2].t = (double)(tex->tex->y) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[1].s = (double)(tex->tex->x) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[1].t = (double)(tex->tex->y) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[0].s = (double)(tex->tex->x) / TEXTURE_COORDINATE_SCALE;
   SpriteTex[0].t = (double)(tex->tex->y + tex->tex->Height) / TEXTURE_COORDINATE_SCALE;


   glPushMatrix();
   glTranslated(x, y, z);
   glRotated(tr2_Theta * 180.0 / Pi, 0, 1, 0);  // make it always face the camera

   glEnableClientState(GL_TEXTURE_COORD_ARRAY);
   glEnable(GL_TEXTURE_2D);

   glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), &SpriteMesh);
   glTexCoordPointer(2, GL_FLOAT, sizeof(opengl_texture_coord), &SpriteTex);
   glBindTexture(GL_TEXTURE_2D, tex->TextureID);
   glDrawArrays(GL_QUADS, 0, 4);

   glPopMatrix();

#ifdef NEVER_COMPILE_THIS     // this works just as well, but it's S-L-O-W
   opengl_texture *tex;
   GLfloat ProximityScale;
   bitu32 TextureSize;
   static bitu32 tdata_size = 0;
   static GLubyte *tdata = NULL;

   tex = opengl_item.SpriteTextures[texture];
   ProximityScale = Distance((int)opengl_camera.eye.x, (int)opengl_camera.eye.y, (int)opengl_camera.eye.z, (int)x, (int)y, (int)z);
   // As the camera approaches the sprite, this can grow quite large.  Here we
   // make sure it doesn't hit 0, and also prevent _extreme_ slowdowns...
   if (ProximityScale < 1000.0)
      ProximityScale = 1000.0;
   ProximityScale = scale / ProximityScale;

   // try to minimise memory fragmentation by re-using our buffer whenever possible
   TextureSize = tex->w * tex->h * ProximityScale * ProximityScale * 4; // the *4 is for R/G/B/A bytes
   if (TextureSize > tdata_size) {
      if (tdata != NULL)
         free(tdata);
      tdata = TR2alloc(TextureSize);
      tdata_size = TextureSize;
      }

   // the gluScaleImage() call is quite expensive, render-time-wise, but it
   // can't really be helped (or can it?)
   gluScaleImage(GL_RGBA, tex->w, tex->h, GL_UNSIGNED_BYTE, tex->data,
                 tex->w * ProximityScale, tex->h * ProximityScale, GL_UNSIGNED_BYTE, tdata);
   glRasterPos4d(x, y, z, 1.0);
   glDrawPixels(tex->w * ProximityScale, tex->h * ProximityScale, GL_RGBA, GL_UNSIGNED_BYTE, tdata);
#endif
}

/*
 * OpenGL callback - this is where things actually get drawn.
 */
void opengl_draw(void)
{
   int i,
       j,
       k,
       l,
       FrameOffset,
       FrameStep,
       MeshList[MAX_NUM_ROOM_MESHES],
       tmpMeshList[MAX_NUM_ROOM_MESHES],
       numMeshList;
   double FrameRotationX[50],
          FrameRotationY[50],
          FrameRotationZ[50];
   double ItemRelX,
          ItemRelY,
          ItemRelZ;
   double ItemAnimX,
          ItemAnimY,
          ItemAnimZ;
   double ItemAbsX,
          ItemAbsY,
          ItemAbsZ;
   double ItemAbsRotY;

   /* clear screen */
   glDisable(GL_DITHER);                                 // disable dither before glClear for efficiency
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // clear both bits at once for efficiency

   i = -1;  //////////// we really need a viable visibility test...

   if (SingleRoomMode) {
      MeshList[0] = WhichSingleRoom;
      numMeshList = 1;
      }
   else if (i < 0) {   // don't know where we are, so draw everything
      // do a depth-sort from the viewpoint (draw most-distant stuff first) so transparency will work properly
      for (i = 0; i < opengl_level->NumRooms; ++i)
         MeshList[i] = i;
      numMeshList = opengl_level->NumRooms;
      // set up a "virtual camera" some distance behind us
      vTheta = tr2_Theta + Pi;
      if (vTheta > Pi)
         vTheta -= TwoPi;
      vCameraX = (64.0 * sin(vTheta)) + opengl_camera.eye.x;
      vCameraY = opengl_camera.eye.y - 256.0;   // put vCamera above us (for water depth sorting)
      vCameraZ = (64.0 * cos(vTheta)) + opengl_camera.eye.z;
      qsort((void *)MeshList, (size_t)numMeshList, (size_t)sizeof(MeshList[0]), RoomDepthCompare);
      }
   else {   // know where we are, draw only us and neighbours
      numMeshList = opengl_item.RoomMeshes[i]->NumAdjacentRooms;
      for (j = 0; j < numMeshList; ++j) {
         tmpMeshList[j] = opengl_item.RoomMeshes[i]->AdjacentRooms[j];
         }
      // run through neighbour's neighbours
      k = numMeshList;
      for (j = 0; j < numMeshList; ++j) {
         for (l = 0; l < opengl_item.RoomMeshes[tmpMeshList[j]]->NumAdjacentRooms; ++l) {
            tmpMeshList[k++] = opengl_item.RoomMeshes[tmpMeshList[j]]->AdjacentRooms[l];
            }
         }
      for (i = 0, numMeshList = 0; i < k; ++i) {
         for (j = 0; j < numMeshList; ++j) {
            if (MeshList[j] == tmpMeshList[i]) // already there
               break;
            }
         if (j == numMeshList) {
            MeshList[numMeshList++] = tmpMeshList[i];
            }
         }
      }

   /* adjust point-of-view */
   opengl_camera.target.z = (tr2_ViewDistance * cos(tr2_Theta)) + opengl_camera.eye.z;
   opengl_camera.target.x = (tr2_ViewDistance * sin(tr2_Theta)) + opengl_camera.eye.x;
   opengl_camera.target.y = opengl_camera.eye.y + EyeTargetOffset;

   /* reset modelview matrix */
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   /* set modelview matrix */
   gluLookAt(opengl_camera.eye.x, opengl_camera.eye.y, opengl_camera.eye.z,
             opengl_camera.target.x, opengl_camera.target.y, opengl_camera.target.z,
             0.0, -1.0, 0.0);

   glGetDoublev(GL_MODELVIEW_MATRIX, ModelViewMatrix);

   /* set ambient colour to white */
   glColor3fv(White);

   WhichLight = 0;
   if (LightingMode) {
      CurrentRoom = WhichRoomCamera();
      }
   else
      glDisable(GL_LIGHTING);

   // draw sky
   if ((DrawSkyMeshMode) && SkyMesh >= 0) {  // position?  rotation?
      double SkyScale = 100.0; // just a guess...
      ItemAbsRotY = 0;
      glDisable(GL_DEPTH_TEST);
      glPushMatrix();
      if (opengl_level->EngineVersion == TombRaider_2) {
         glRotated(90.0, 1, 0, 0);
         }
      glTranslated(0.0, 1000.0, 0.0);
      glScaled(SkyScale, SkyScale, SkyScale);
      opengl_draw_item_mesh(opengl_item.ItemMeshes[SkyMesh]);
      glPopMatrix();
      glEnable(GL_DEPTH_TEST);
      }

   if (BoxesMode) {
      const GLfloat BoxColours[8][4] = {
         { 0.8,  0.8,  0.8, 0.8 },  // white
         { 1.0,  0.0,  0.0, 0.8 },  // red
         { 1.0,  1.0,  0.0, 0.8 },  // yellow
         { 0.0,  1.0,  0.0, 0.8 },  // green
         { 0.0,  1.0,  1.0, 0.8 },  // cyan
         { 0.5,  0.5,  0.5, 0.8 },  // grey
         { 0.0,  0.0,  1.0, 0.8 },  // blue
         { 1.0,  0.0,  1.0, 0.8 }   // magenta
         };

      for (i = 0; i < (int)opengl_level->NumBoxes; ++i) {
         opengl_vector Box[4];
         int i8 = (int)opengl_level->NumBoxes / 8;
         bitu8 *cptr = (bitu8 *)&opengl_level->Boxes[i * 4];

         Box[0].x = (bitu16)opengl_level->Boxes[i].Xmin * 1024;
         Box[0].y = (bit16)opengl_level->Boxes[i].TrueFloor;
         Box[0].z = (bitu16)opengl_level->Boxes[i].Zmin * 1024;

         Box[3].x = (bitu16)opengl_level->Boxes[i].Xmax * 1024;
         Box[3].y = (bit16)opengl_level->Boxes[i].TrueFloor;
         Box[3].z = (bitu16)opengl_level->Boxes[i].Zmin * 1024;
         
         Box[2].x = (bitu16)opengl_level->Boxes[i].Xmax * 1024;
         Box[2].y = (bit16)opengl_level->Boxes[i].TrueFloor;
         Box[2].z = (bitu16)opengl_level->Boxes[i].Zmax * 1024;
         
         Box[1].x = (bitu16)opengl_level->Boxes[i].Xmin * 1024;
         Box[1].y = (bit16)opengl_level->Boxes[i].TrueFloor;
         Box[1].z = (bitu16)opengl_level->Boxes[i].Zmax * 1024;

         glDisableClientState(GL_TEXTURE_COORD_ARRAY);
         glDisableClientState(GL_NORMAL_ARRAY);
         glDisable(GL_TEXTURE_2D);

         glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), Box);
         glColor4fv(BoxColours[i % 8]);
         glDrawArrays(GL_QUADS, 0, 4);
         }
      }

   if (++AnimatedTextureTick > 99)
      AnimatedTextureTick = 0;

   // draw room meshes
   glEnableClientState(GL_TEXTURE_COORD_ARRAY);
   glEnable(GL_TEXTURE_2D);
   for (i = 0; i < numMeshList; ++i) {
      opengl_draw_room_mesh(opengl_item.RoomMeshes[MeshList[i]]);
      }

   // draw static meshes (if enabled)
   if (DrawStaticMeshes) {
      for (i = 0; i < numMeshList; ++i) {
         for (j = 0; j < opengl_level->Rooms[MeshList[i]].NumStaticMeshes; ++j) {
            for (k = 0; k < (int)opengl_level->NumStaticMeshes; ++k) {
               if (opengl_level->Rooms[MeshList[i]].StaticMeshes[j].ObjectID == opengl_level->StaticMeshes[k].ObjectID) {
                  ItemAbsX = opengl_level->Rooms[MeshList[i]].StaticMeshes[j].x;
                  ItemAbsY = opengl_level->Rooms[MeshList[i]].StaticMeshes[j].y;
                  ItemAbsZ = opengl_level->Rooms[MeshList[i]].StaticMeshes[j].z;
                  if (IsBehind((int)ItemAbsX, (int)ItemAbsZ))  // don't bother drawing things that are behind the camera
                     break;
                  ItemAbsRotY = ((opengl_level->Rooms[MeshList[i]].StaticMeshes[j].Rotation >> 14) & 0x03) * 90;
                  glPushMatrix();
                  glTranslated(ItemAbsX, ItemAbsY, ItemAbsZ);
                  glRotated(ItemAbsRotY, 0, 1, 0);
                  opengl_draw_item_mesh(opengl_item.ItemMeshes[opengl_level->StaticMeshes[k].StartingMesh]);
                  glPopMatrix();
                  break;
                  }
               }
            }
         }
      }

   // draw room sprites (if enabled)
   if (DrawRoomSprites) {
      for (i = 0; i < numMeshList; ++i) {
         for (j = 0; j < opengl_level->Rooms[MeshList[i]].RoomData.NumSprites; ++j) {
            tr2_vertex *sVertex;
            sVertex = &opengl_level->Rooms[MeshList[i]].RoomData.Vertices[opengl_level->Rooms[MeshList[i]].RoomData.Sprites[j].Vertex].Vertex;
            if (IsBehind(opengl_level->Rooms[MeshList[i]].info.x + sVertex->x, opengl_level->Rooms[MeshList[i]].info.z + sVertex->z))   // don't bother drawing things that are behind the camera
               continue;
            opengl_draw_sprite(opengl_level->Rooms[MeshList[i]].info.x + sVertex->x, sVertex->y, opengl_level->Rooms[MeshList[i]].info.z + sVertex->z, opengl_level->Rooms[MeshList[i]].RoomData.Sprites[j].Texture, 10.0);
            }
         }
      }

   // draw item meshes (if enabled)
   if (MeshMode) {
      for (j = opengl_level->NumItems - 1; j >= 0; --j) {
         // Determine world-coordinate position and rotation from Items[]
         ItemAbsX = opengl_level->Items[j].x;
         ItemAbsY = opengl_level->Items[j].y;
         ItemAbsZ = opengl_level->Items[j].z;
         if (IsBehind((int)ItemAbsX, (int)ItemAbsZ)) {   // don't bother drawing things that are completely behind the camera
            continue;
            }
         ItemAbsRotY = ((opengl_level->Items[j].Angle >> 14) & 0x03);
         ItemAbsRotY *= 90.0;
         k = opengl_level->Items[j].ObjectID;
         //
         // At this point, we scan Sprites and Moveables for the current ObjectID.
         // We scan Sprites first, because TR1 actually re-uses ObjectIDs (for
         // example, in LEVEL1.PHD, ObjectID 150 is both a large medi-pack and a
         // piece of the Scion).
         //
         i = 0;   // "Could be a Sprite" flag (0 == "could be a sprite")
         if (opengl_level->EngineVersion == TombRaider_1 &&
             opengl_level->Items[j].Intensity1 == -1) {  // it's a mesh, not a sprite
            i = (int)opengl_level->NumSpriteSequences;   // skip the "for" loop that follows
            }
         // search the SpriteSequence list (if we didn't already decide that it's a mesh)
         for ( ; i < (int)opengl_level->NumSpriteSequences; ++i) {
            if (opengl_level->SpriteSequences[i].ObjectID == k)
               break;
            }
         if (i == (int)opengl_level->NumSpriteSequences) {  // not in SpriteSequences[] (or we know it's not a sprite)
            // search the Moveable list
            for (i = 0; i < (int)opengl_level->NumMoveables; ++i)
               if ((int)opengl_level->Moveables[i].ObjectID == k) {
                  break;
                  }
            if (i == (int)opengl_level->NumMoveables) {     // not in SpriteSequences[] _or_ Moveables[]
               i = 0;            // Lara (easily noticed)
               ItemAbsY -= 470;  // KLUDGE - adjust for height
               }
            }
         else {   // it's a sprite
            opengl_draw_sprite(ItemAbsX, ItemAbsY, ItemAbsZ, opengl_level->SpriteSequences[i].Offset, 4.0);
            continue;
            }
         k = i;   // k == which moveable

         // choose an animation with which to pose the moveable
         i = opengl_level->Moveables[k].Animation;
         // If there is more than one, choose the second one (tends to be more interesting).
         // Without going into the state data, it's hard to choose a single animation that
         // is interesting for most of the moveables.  Making Lara, Tigers, and Spiders run
         // also makes doors, drop-floors, and crumbly floors sit perfectly still;  making
         // the doors move tends to make the spiders and tigers go into an intermediate
         // animation where they just twitch.  Oh, well...  that goes on the TO-DO list for V2.0...
         if (k != ((int)opengl_level->NumMoveables - 1)) {
            if (opengl_level->Moveables[k + 1].Animation != (opengl_level->Moveables[k].Animation + 1))
               ++i;
            }
         // calculate index into Frames[] and how large the frame is
         FrameOffset = opengl_level->Animations[i].FrameOffset / 2;
         FrameStep = opengl_level->Animations[i].FrameSize;
         // choose the next frame to draw, based upon FrameCycle
         ++i;
         if (i >= (int)opengl_level->NumAnimations)
            i = opengl_level->NumFrames - FrameOffset;
         else
            i = (opengl_level->Animations[i].FrameOffset / 2) - FrameOffset;
         if (FrameStep != 0)  // prevent divide-by-zero errors
            i /= FrameStep;
         if (i != 0) // prevent divide-by-zero errors
            FrameOffset += FrameStep * (FrameCycle % i);

         // TR3 KLUDGE: Lara's ObjectID is 315, but her meshes start at 0, so make a
         // quick substitution (so she doesn't appear as a bunch of thighs)
         if (k == 0 && opengl_level->EngineVersion == TombRaider_3) {   // fix Lara in TR3
            int zz;
            for (zz = 0; zz < (int)opengl_level->NumMoveables && k == 0; ++zz) {
               if (opengl_level->Moveables[zz].ObjectID == 315)
                  k = zz;
               }
            }
TR4loop:;   // Uh-oh... "GOTO" labels are never a good sign...

         // TR4 KLUDGE: Lara is "skinned" in TR4, which seems to mean that she has two sets
         // of meshes - one for her "main" body, the other for her joints (knees, etc.).  Here
         // we do two substitutions - the first draws the main body, then after the rendering,
         // we see if we just rendered the main body - if so, then we loop back up here, where
         // we make a second substitution and draw the joints.  It's imperfect, but close
         // enough for this viewer...
         if (k == 0 && opengl_level->EngineVersion == TombRaider_4) {   // fix Lara in TR4
            int zz;
            for (zz = 0; zz < (int)opengl_level->NumMoveables && k == 0; ++zz) {
               if (opengl_level->Moveables[zz].ObjectID == 8)   // body is ItemID 8, joints are ItemID 9 (TR4 demo: body is ItemID 10, joints are ItemID 11)
                  k = zz;
               }
            }
         else if (opengl_level->Moveables[k].ObjectID == 8 && opengl_level->EngineVersion == TombRaider_4) {  // KLUDGE to do "skinning"
            int zz;
            k = 0;
            for (zz = 0; zz < (int)opengl_level->NumMoveables && k == 0; ++zz) {
               if (opengl_level->Moveables[zz].ObjectID == 9)   // body is ItemID 8, joints are ItemID 9 (TR4 demo: body is ItemID 10, joints are ItemID 11)
                  k = zz;
               }
            }
         //
         // OK, at this point, <k> indicates which Moveable we're using, and
         // <FrameOffset> is pointed into Frames[] at the appropriate spot for
         // this moveable at this step in its animation.
         //
         // Figure out where it should be positioned, based upon the frame:
         //
         ItemAnimX = (bit16)opengl_level->Frames[FrameOffset + 6];
         ItemAnimY = (bit16)opengl_level->Frames[FrameOffset + 7];
         ItemAnimZ = (bit16)opengl_level->Frames[FrameOffset + 8];
         glPushMatrix();
         //
         // For efficiency, we jump ahead a bit here, and calculate an array of
         // rotations from Frames[FrameOffset] (so that the main rendering loop
         // can just use them instead of calculating them on the fly).
         //
         for (i = 0; i < 50; ++i) {    // initialise rotation array (size should really be dynamic)
            FrameRotationX[i] = 0.0;
            FrameRotationY[i] = 0.0;
            FrameRotationZ[i] = 0.0;
            }
         l = 9;   // first angle offset in this Frame
         // run through all the meshes and calculate the rotation angles
         for (i = 0; i < opengl_level->Moveables[k].NumMeshes; ++i) { // get actual rotations
            bitu16 itmp = opengl_level->Frames[FrameOffset + l++],
                   itmp2;
            double angle;

            if (opengl_level->EngineVersion == TombRaider_1) {
               // all angles are three-axis
               angle = (itmp >> 4) & 0x03ff;
               angle *= 360.0 / 1024.0;
               FrameRotationX[i] = angle;

               itmp2 = (itmp << 6) & 0x03c0;
               itmp = opengl_level->Frames[FrameOffset + (l /* - 1 */)]; // get Z rotation
               ++l;

               itmp2 |= (itmp >> 10) & 0x003f;
               angle = itmp2;
               angle *= 360.0 / 1024.0;
               FrameRotationY[i] = angle;

               angle = itmp & 0x3ff;
               angle *= 360.0 / 1024.0;
               FrameRotationZ[i] = angle;
               }
            else if (itmp & 0xc000) { // TR2, TR3, TR4 - single axis of rotation
               if (opengl_level->EngineVersion == TombRaider_4) {
                  angle = itmp & 0x0fff;
                  angle /= 4096.0;
                  angle *= 360.0;
                  }
               else {
                  angle = itmp & 0x3ff;
                  angle /= 1024.0;
                  angle *= 360.0;
                  }
               switch (itmp & 0xc000) {
                  case 0x4000 :
                     FrameRotationX[i] = angle;
                     break;
                  case 0x8000 :
                     FrameRotationY[i] = angle;
                     break;
                  case 0xc000 :
                     FrameRotationZ[i] = angle;
                     break;
                  }
               }
            else {   // TR2, TR3, TR4 - three axes
               angle = (itmp >> 4) & 0x03ff;
               angle *= 360.0 / 1024.0;
               FrameRotationX[i] = angle;

               itmp2 = (itmp << 6) & 0x03c0;
               itmp = opengl_level->Frames[FrameOffset + l++]; // get Z rotation

               itmp2 |= (itmp >> 10) & 0x003f;
               angle = itmp2;
               angle *= 360.0 / 1024.0;
               FrameRotationY[i] = angle;

               angle = itmp & 0x3ff;
               angle *= 360.0 / 1024.0;
               FrameRotationZ[i] = angle;
               }
            }
         //
         // Now draw all of the meshes, positioned and rotated appropriately.
         // We get the relative positions from MeshTrees[].
         //
         for (i = 0; i < opengl_level->Moveables[k].NumMeshes; ++i) {
            int ThisMesh = opengl_level->Moveables[k].StartingMesh + i;

            if (i == 0) {  // first mesh - position to world coordinates, set rotation
               glTranslated(ItemAbsX, ItemAbsY, ItemAbsZ);
               glRotated(ItemAbsRotY, 0, 1, 0);
               glTranslated(ItemAnimX, ItemAnimY, ItemAnimZ);
               ItemRelX = 0;
               ItemRelY = 0;
               ItemRelZ = 0;
               }
            else {   // (i > 0): second+ meshes - position relative to first mesh (from MeshTree)
               bit32 *itmp;
               tr2_meshtree *MeshTree;

               // KLUDGE of sorts - Moveables[k].MeshTree is a byte offset into MeshTrees[],
               // so we have to do a little converting here...
               itmp = (bit32 *)opengl_level->MeshTrees;
               MeshTree = (tr2_meshtree *)&itmp[opengl_level->Moveables[k].MeshTree + ((i - 1) * 4)];

               ItemRelX = MeshTree->x;
               ItemRelY = MeshTree->y;
               ItemRelZ = MeshTree->z;
               if (MeshTree->Flags & 0x01) { // return to last anchor
                  glPopMatrix();
                  }
               if (MeshTree->Flags & 0x02) { // make new anchor
                  glPushMatrix();
                  }
               }

            // translate, rotate, and render.
            glTranslated(ItemRelX, ItemRelY, ItemRelZ);
            if (FrameRotationY[i])
               glRotated(FrameRotationY[i], 0, 1, 0);
            if (FrameRotationX[i])
               glRotated(FrameRotationX[i], 1, 0, 0);
            if (FrameRotationZ[i])
               glRotated(FrameRotationZ[i], 0, 0, 1);
            opengl_draw_item_mesh(opengl_item.ItemMeshes[ThisMesh]);
            } // end for (i = 0; i < opengl_level->Moveables[k].NumMeshes...
         glPopMatrix();
         // TR4 KLUDGE: if this is Lara, and we've just drawn her extremities, loop again to draw her joints
         if (opengl_level->Moveables[k].ObjectID == 8 && opengl_level->EngineVersion == TombRaider_4)   // KLUDGE to do "skinning"
            goto TR4loop;
         }  // end for (j = opengl_level->NumItems - 1; j >= 0...
      }  // end if (MeshMode)


   // update coordinate display & reset message buffer
   opengl_print(MessageBuffer);

   /* update physical display */
   glutSwapBuffers();

   // make the redisplay an infinite loop
   glutPostRedisplay();

   // If we're animating the meshes, advance the counter so we'll choose the next
   // frame the next time through.
   if (AnimateMeshes)
      ++FrameCycle;
}

/*
 * Draw a room mesh.
 */
void opengl_draw_room_mesh(opengl_mesh *mesh)
{
   int i,
       ThisTexture,
       LastTexture;

   if (IsBehind((int)mesh->BoundingBoxLow.x, (int)mesh->BoundingBoxLow.z) &&
       IsBehind((int)mesh->BoundingBoxHigh.x, (int)mesh->BoundingBoxHigh.z) &&
       IsBehind((int)mesh->BoundingBoxHigh.x, (int)mesh->BoundingBoxLow.z) &&
       IsBehind((int)mesh->BoundingBoxLow.x, (int)mesh->BoundingBoxHigh.z)) {
      // don't bother with rooms that are completely behind the camera
      return;
      }

   /* set up matrix */
   glPushMatrix();

   Setup_Lighting();

   // position the room to its base coordinates
   glTranslated(mesh->Position.x, mesh->Position.y, mesh->Position.z);

   /* draw texture mapped rectangles */
   if (mesh->oglNumTexturedRectangles > 0) {
      if (DoPerVertexLighting && mesh->oglRectangleVertexColours != NULL) {
         if (!ColourArrayEnabled) {
            glEnableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = TRUE;
            }
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglRectangleVertexColours);
         }
      else {
         if (ColourArrayEnabled) {
            glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = FALSE;
            }
         }
      if (AnimatedTextureMode) {
         // we're guaranteed that opengl_level->NumAnimatedTextures != 0 && this is not TR4...
         if ((AnimatedTextureTick % AnimatedTextureCycle) == 0) { // slow things down a bit...
            int ThisRoom = mesh->AdjacentRooms[0];
            int NextTexture,
                NumTextures,
                base,
                j,
                k,
                t;
            for (i = 0, base = 1; i < opengl_level->AnimatedTextures[0]; ++i) {
               NumTextures = opengl_level->AnimatedTextures[base++] + 1;
               for (j = NumTextures - 1; j >= 0; --j) {  // scan the textures backwards so we don't change them more than once (this doesn't work if there are only two textures!)
                  for (k = 0; k < mesh->oglNumTexturedRectangles; ++k) {
                     if (opengl_level->AnimatedTextures[base + j] == opengl_level->Rooms[ThisRoom].RoomData.Rectangles[k].Texture) {
                        NextTexture = opengl_level->AnimatedTextures[base + ((j + 1) % NumTextures)];
                        opengl_level->Rooms[ThisRoom].RoomData.Rectangles[k].Texture = NextTexture;
                        mesh->oglRectangleTextures[k]->TextureID = TexTileIDs[opengl_level->ObjectTextures[NextTexture].Tile];
                        for (t = 0; t < 4; ++t) {
                           mesh->oglRectangleTextureCoordinates[(k * 4) + t].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[NextTexture].Vertices[t].Xpixel, opengl_level->ObjectTextures[NextTexture].Vertices[t].Xcoordinate);
                           mesh->oglRectangleTextureCoordinates[(k * 4) + t].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[NextTexture].Vertices[t].Ypixel, opengl_level->ObjectTextures[NextTexture].Vertices[t].Ycoordinate);
                           }
                        }
                     }
                  }
               base += NumTextures;
               }
            }
         }
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedRectangleVertices);
      glTexCoordPointer(2, GL_FLOAT, sizeof(opengl_texture_coord), mesh->oglRectangleTextureCoordinates);
      for (i = 0, LastTexture = -1; i < mesh->oglNumTexturedRectangles; ) {
         int j;
         //
         // For efficiency, we sorted the polygons by TextureID;  this way, we can render
         // all of the polygons of a given texture at one time, thus avoiding the overhead
         // of re-binding the texture for every polygon.
         //
         ThisTexture = mesh->oglRectangleTextures[i]->TextureID;
         if (ThisTexture != LastTexture) {
            glBindTexture(GL_TEXTURE_2D, ThisTexture);
            LastTexture = ThisTexture;
            }
         // find the next texture change
         for (j = 1; (i + j) < mesh->oglNumTexturedRectangles; ++j) {
            if (ThisTexture != (int)mesh->oglRectangleTextures[i + j]->TextureID)
               break;
            }
         glDrawArrays(GL_QUADS, i * 4, 4 * j);
         i += j;
         }
      }

   /* draw texture mapped triangles */
   if (mesh->oglNumTexturedTriangles > 0) {
      if (DoPerVertexLighting && mesh->oglTriangleVertexColours != NULL) {
         if (!ColourArrayEnabled) {
            glEnableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = TRUE;
            }
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglTriangleVertexColours);
         }
      else {
         if (ColourArrayEnabled) {
            glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = FALSE;
            }
         }
      if (AnimatedTextureMode) {
         // we're guaranteed that opengl_level->NumAnimatedTextures != 0 && this is not TR4...
         if ((AnimatedTextureTick % AnimatedTextureCycle) == 0) { // slow things down a bit...
            int ThisRoom = mesh->AdjacentRooms[0];
            int NextTexture,
                NumTextures,
                base,
                j,
                k,
                t;
            for (i = 0, base = 1; i < opengl_level->AnimatedTextures[0]; ++i) {
               NumTextures = opengl_level->AnimatedTextures[base++] + 1;
               for (j = NumTextures - 1; j >= 0; --j) {  // scan the textures backwards so we don't change them more than once (this doesn't work if there are only two textures!)
                  for (k = 0; k < mesh->oglNumTexturedTriangles; ++k) {
                     if (opengl_level->AnimatedTextures[base + j] == opengl_level->Rooms[ThisRoom].RoomData.Triangles[k].Texture) {
                        NextTexture = opengl_level->AnimatedTextures[base + ((j + 1) % NumTextures)];
                        opengl_level->Rooms[ThisRoom].RoomData.Triangles[k].Texture = NextTexture;
                        mesh->oglTriangleTextures[k]->TextureID = TexTileIDs[opengl_level->ObjectTextures[NextTexture].Tile];
                        for (t = 0; t < 3; ++t) {
                           mesh->oglTriangleTextureCoordinates[(k * 3) + t].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[NextTexture].Vertices[t].Xpixel, opengl_level->ObjectTextures[NextTexture].Vertices[t].Xcoordinate);
                           mesh->oglTriangleTextureCoordinates[(k * 3) + t].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[NextTexture].Vertices[t].Ypixel, opengl_level->ObjectTextures[NextTexture].Vertices[t].Ycoordinate);
                           }
                        }
                     }
                  }
               base += NumTextures;
               }
            }
         }
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedTriangleVertices);
      glTexCoordPointer(2, GL_FLOAT, sizeof(opengl_texture_coord), mesh->oglTriangleTextureCoordinates);
      for (i = 0, LastTexture = -1; i < mesh->oglNumTexturedTriangles; ) {
         int j;
         ThisTexture = mesh->oglTriangleTextures[i]->TextureID;
         //
         // For efficiency, we sorted the polygons by TextureID;  this way, we can render
         // all of the polygons of a given texture at one time, thus avoiding the overhead
         // of re-binding the texture for every polygon.
         //
         if (ThisTexture != LastTexture) {
            glBindTexture(GL_TEXTURE_2D, ThisTexture);
            LastTexture = ThisTexture;
            }
         // find the next texture change
         for (j = 1; (i + j) < mesh->oglNumTexturedTriangles; ++j) {
            if (ThisTexture != (int)mesh->oglTriangleTextures[i + j]->TextureID)
               break;
            }
         glDrawArrays(GL_TRIANGLES, i * 3, 3 * j);
         i += j;
         }
      }
//   glDisableClientState(GL_COLOR_ARRAY);
//   ColourArrayEnabled = FALSE;

   /*
    * If desired, draw portals using coloured rectangles.
    */
   if (PortalMode) {
      /* set up for plain polygons */
      glDisable(GL_TEXTURE_2D);

      /* draw plain rectangles */
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredRectangleVertices);
      if (LightingMode) {
         glNormalPointer(GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredRectangleNormals);
         glEnableClientState(GL_NORMAL_ARRAY);
         }
      for (i = 0; i < mesh->oglNumColouredRectangles; ++i) {
         glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, (GLfloat *)&mesh->oglRectangleColours[i]);
         glDrawArrays(GL_QUADS, i * 4, 4);
         }
      if (LightingMode)
         glDisableClientState(GL_NORMAL_ARRAY);
      glEnable(GL_TEXTURE_2D);
      }

   /* reset matrix */
   glPopMatrix();

//   glColor3fv(White);
}

/*
 * Draw an item mesh (e.g. Lara, boulders, bad guys, tigers, etc.)
 */
void opengl_draw_item_mesh(opengl_mesh *mesh) // this gets called a LOT, needs optimisation
{
   int i,
       ThisTexture,
       LastTexture;

   if (mesh == NULL) // sanity check
      return;

   if (mesh->oglNumTexturedRectangles > 0 || mesh->oglNumTexturedTriangles > 0) {
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glEnable(GL_TEXTURE_2D);
      }

   glColor3fv(White);   // ensure that lighting is white and not some random colour
   /* draw texture mapped rectangles */
   if (mesh->oglNumTexturedRectangles > 0) {
      if (mesh->oglRectangleVertexColours != NULL) {
         if (DoPerVertexLighting) {
            if (!ColourArrayEnabled) {
               glEnableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
               ColourArrayEnabled = TRUE;
               }
            }
         else {
            if (ColourArrayEnabled) {
               glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
               ColourArrayEnabled = FALSE;
               }
            }
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglRectangleVertexColours);
         }
      else {
         if (ColourArrayEnabled) {
            glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = FALSE;
            }
         glEnableClientState(GL_NORMAL_ARRAY);
         }
      glNormalPointer(GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedRectangleNormals);
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedRectangleVertices);
      glTexCoordPointer(2, GL_FLOAT, sizeof(opengl_texture_coord), mesh->oglRectangleTextureCoordinates);
      for (i = 0, LastTexture = -1; i < mesh->oglNumTexturedRectangles; ) {
         int j;
         //
         // For efficiency, we sorted the polygons by TextureID;  this way, we can render
         // all of the polygons of a given texture at one time, thus avoiding the overhead
         // of re-binding the texture for every polygon.
         //
         ThisTexture = mesh->oglRectangleTextures[i]->TextureID;
         if (ThisTexture != LastTexture) {
            glBindTexture(GL_TEXTURE_2D, ThisTexture);
            LastTexture = ThisTexture;
            }
         // find the next texture change
         for (j = 1; (i + j) < mesh->oglNumTexturedRectangles; ++j) {
            if (ThisTexture != (int)mesh->oglRectangleTextures[i + j]->TextureID)
               break;
            }
         glDrawArrays(GL_QUADS, i * 4, 4 * j);
         i += j;
         }
      }
   glDisableClientState(GL_NORMAL_ARRAY);

   /* draw texture mapped triangles */
   if (mesh->oglNumTexturedTriangles > 0) {
      if (mesh->oglTriangleVertexColours != NULL) {
         if (DoPerVertexLighting) {
            if (!ColourArrayEnabled) {
               glEnableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
               ColourArrayEnabled = TRUE;
               }
            }
         else {
            if (ColourArrayEnabled) {
               glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
               ColourArrayEnabled = FALSE;
               }
            }
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglTriangleVertexColours);
         }
      else {
         if (ColourArrayEnabled) {
            glDisableClientState(GL_COLOR_ARRAY);   // per-vertex lighting
            ColourArrayEnabled = FALSE;
            }
         glEnableClientState(GL_NORMAL_ARRAY);
         }
      glNormalPointer(GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedTriangleNormals);
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglTexturedTriangleVertices);
      glTexCoordPointer(2, GL_FLOAT, sizeof(opengl_texture_coord), mesh->oglTriangleTextureCoordinates);
      for (i = 0, LastTexture = -1; i < mesh->oglNumTexturedTriangles; ) {
         int j;
         //
         // For efficiency, we sorted the polygons by TextureID;  this way, we can render
         // all of the polygons of a given texture at one time, thus avoiding the overhead
         // of re-binding the texture for every polygon.
         //
         ThisTexture = mesh->oglTriangleTextures[i]->TextureID;
         if (ThisTexture != LastTexture) {
            glBindTexture(GL_TEXTURE_2D, ThisTexture);
            LastTexture = ThisTexture;
            }
         // find the next texture change
         for (j = 1; (i + j) < mesh->oglNumTexturedTriangles; ++j) {
            if (ThisTexture != (int)mesh->oglTriangleTextures[i + j]->TextureID)
               break;
            }
         glDrawArrays(GL_TRIANGLES, i * 3, 3 * j);
         i += j;
         }
      }
   glDisableClientState(GL_NORMAL_ARRAY);

   /* set up for plain polygons */
   if (mesh->oglNumColouredRectangles > 0 || mesh->oglNumColouredTriangles > 0) {
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glDisable(GL_TEXTURE_2D);
      }
   glDisableClientState(GL_COLOR_ARRAY);
   ColourArrayEnabled = FALSE;

   /* draw coloured rectangles */
   if (mesh->oglNumColouredRectangles > 0) {
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredRectangleVertices);
      if (mesh->oglCRectangleVertexColours != NULL) {
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglCRectangleVertexColours);
         glDisableClientState(GL_NORMAL_ARRAY);
         }
      else {
         glNormalPointer(GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredRectangleNormals);
         glEnableClientState(GL_NORMAL_ARRAY);
         }
      for (i = 0; i < mesh->oglNumColouredRectangles; ++i) {
         glColor4fv((GLfloat *)&mesh->oglRectangleColours[i]);
         glDrawArrays(GL_QUADS, i * 4, 4);
         }
      }

   /* draw coloured triangles */
   if (mesh->oglNumColouredTriangles > 0) {
      glVertexPointer(3, GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredTriangleVertices);
      if (mesh->oglCTriangleVertexColours != NULL) {
         glColorPointer(4, GL_FLOAT, sizeof(opengl_colour), (GLfloat *)mesh->oglCTriangleVertexColours);
         glDisableClientState(GL_NORMAL_ARRAY);
         }
      else {
         glNormalPointer(GL_FLOAT, sizeof(opengl_vector), mesh->oglColouredTriangleNormals);
         glEnableClientState(GL_NORMAL_ARRAY);
         }
      for (i = 0; i < mesh->oglNumColouredTriangles; ++i) {
         glColor4fv((GLfloat *)&mesh->oglTriangleColours[i]);
         glDrawArrays(GL_TRIANGLES, i * 3, 3);
         }
      }

   if (mesh->oglNumColouredRectangles > 0 ||
       mesh->oglNumColouredTriangles > 0 ||
       ColourArrayEnabled) {  // DoPerVertexLighting
      glDisableClientState(GL_COLOR_ARRAY);
      ColourArrayEnabled = FALSE;
      glColor3fv(White);
      }
}

