#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/timeb.h>

#include "config.h"
#include "tr2.h"

#ifdef INTELPC
/*
 * The #define WIN32_LEAN_AND_MEAN and #include <windows.h> are required for
 * gl.h, glu.h, and glut.h - need to make this more portable...
 */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include "texture.h"
#define OPENGL_FULL                 // tell opengl.h to do all the defines and enums that harness.c doesn't want to see
#include "opengl.h"
#include "trmath.h"

/*
 * Macros/Constants
 */
#define ROOM_TRANSPARENCY  0xff  // scale 0 == transparent, 0xFF == opaque

/*
 * Set/retrieve texture information
 */
texture *GetTextureInfo(TR2_Level *Level, TextureType Type, int TextureNumber, TexturePolygonType PolygonType)
{
   texture *tex;
   int NumVertices;
   int Xmin,
       Ymin,
       Xmax,
       Ymax;
   int i;

   /* alloc space for texture details */
   tex = (texture *)malloc(sizeof(texture));

   /* set texture identifiers */
   tex->Type = Type;
   tex->PolygonType = PolygonType;
   tex->TextureNumber = TextureNumber;

   /* object texture type */
   if (Type == OBJECT) {
      /* set data pointer */
      tex->Data = Level->Textile8[Level->ObjectTextures[TextureNumber].Tile].Tile;
      tex->Data16 = Level->Textile16[Level->ObjectTextures[TextureNumber].Tile].Tile;

      /* set palette */
      tex->Palette = Level->Palette8;
    
      /* set dimensions */
      if (PolygonType == TRIANGLE)
         NumVertices = 3;
      else
         NumVertices = 4;
      Xmin = 256;
      Ymin = 256;
      Xmax = 0;
      Ymax = 0;
      for (i = 0; i < NumVertices; i++) {
         if (Level->ObjectTextures[TextureNumber].Vertices[i].Xpixel > Xmax)
            Xmax = Level->ObjectTextures[TextureNumber].Vertices[i].Xpixel;
         if (Level->ObjectTextures[TextureNumber].Vertices[i].Ypixel > Ymax)
            Ymax = Level->ObjectTextures[TextureNumber].Vertices[i].Ypixel;
         if (Level->ObjectTextures[TextureNumber].Vertices[i].Xpixel < Xmin)
            Xmin = Level->ObjectTextures[TextureNumber].Vertices[i].Xpixel;
         if (Level->ObjectTextures[TextureNumber].Vertices[i].Ypixel < Ymin)
            Ymin = Level->ObjectTextures[TextureNumber].Vertices[i].Ypixel;
         }
      tex->Width  = (Xmax - Xmin) + 1;
      tex->Height = (Ymax - Ymin) + 1;
      tex->x = Xmin;
      tex->y = Ymin;
    
      /* return details */
      return(tex);
      }
   else if (Type == SPRITE) { // handle sprite textures
      tr2_sprite_texture *Sprite;

      //
      // In the SPRITE case, "TextureNumber" is actually an offset into SpriteTextures;
      // the resulting "TextureNumber" is NumObjectTextures + TextureNumber
      //
      Sprite = &Level->SpriteTextures[TextureNumber];
      tex->TextureNumber = TextureNumber + Level->NumObjectTextures;
      /* set data pointer */
      tex->Data = Level->Textile8[Sprite->Tile].Tile;
      tex->Data16 = Level->Textile16[Sprite->Tile].Tile;

      /* set palette */
      tex->Palette = Level->Palette8;
    
      /* set dimensions */
      tex->Width  = Sprite->Width >> 8;
      tex->Height = Sprite->Height >> 8;
      tex->x = Sprite->x;
      tex->y = Sprite->y;
    
      /* return details */
      return(tex);
      }
   /* unknown type */
   else {
      fprintf(stderr, "Unknown Texture Type (%d), aborting...\n", Type);
      exit(99);
      }
}

/*
 * Add a texture to the texture list.
 */
opengl_texture *opengl_add_texture(TextureType type, int TextureNumber, TexturePolygonType polygon_type)
{
   int i,
       j;
   opengl_texture *tex;
   int c_index,
       offset;
   static TexTilesInitialised = FALSE;

   // The first time through, we create 256*256 textures from the appropriate textiles
   if (!TexTilesInitialised) {
      TexTileIDs = (GLuint *)TR2alloc(sizeof(GLuint) * opengl_level->NumTextiles);
      glGenTextures(opengl_level->NumTextiles, TexTileIDs);

      for (i = 0; i < (int)opengl_level->NumTextiles; ++i) {
         bitu8 TempTexTile[256 * 256 * 4];
         int k;

         glBindTexture(GL_TEXTURE_2D, TexTileIDs[i]);
         for (j = 0; j < 256; ++j) {
            for (k = 0; k < 256; ++k) {
               c_index = (j * 256) + k;
               offset = opengl_level->Textile16[i].Tile[c_index];
               c_index = (j * 1024) + (k * 4);
               TempTexTile[c_index + 0] = ((offset >> 10) & 0x1f) * 8;
               TempTexTile[c_index + 1] = ((offset >> 5) & 0x1f) * 8;
               TempTexTile[c_index + 2] = (offset & 0x1f) * 8;
               TempTexTile[c_index + 3] = (offset & 0x8000) ? ROOM_TRANSPARENCY : 0;
               }
            }
         // TempTexTile now contains the texture data
         // TR3 KLUDGE: Try to account for its alpha transparencies (rays of light, water surface, fog, etc.)
         if (opengl_level->EngineVersion >= TombRaider_3) {
            for (j = 0; j < (int)opengl_level->NumObjectTextures; ++j) {
               if (opengl_level->ObjectTextures[j].Tile == i &&
                   opengl_level->ObjectTextures[j].TransparencyFlags == 2) {
                  // make this object's pixels transparent
                  int xMin = 999,
                      xMax = 0,
                      yMin = 999,
                      yMax = 0,
                      x,
                      y;
                  y = 4;
                  if (opengl_level->ObjectTextures[j].Vertices[3].Xpixel == 0 &&
                      opengl_level->ObjectTextures[j].Vertices[3].Ypixel == 0)
                     y = 3; // account for triangles
                  for (k = 0; k < y; ++k) {
                     if (opengl_level->ObjectTextures[j].Vertices[k].Xpixel > xMax)
                        xMax = opengl_level->ObjectTextures[j].Vertices[k].Xpixel;
                     if (opengl_level->ObjectTextures[j].Vertices[k].Xpixel < xMin)
                        xMin = opengl_level->ObjectTextures[j].Vertices[k].Xpixel;
                     if (opengl_level->ObjectTextures[j].Vertices[k].Ypixel > yMax)
                        yMax = opengl_level->ObjectTextures[j].Vertices[k].Ypixel;
                     if (opengl_level->ObjectTextures[j].Vertices[k].Ypixel < yMin)
                        yMin = opengl_level->ObjectTextures[j].Vertices[k].Ypixel;
                     }
                  for (x = xMin; x <= xMax; ++x) {
                     for (y = yMin; y <= yMax; ++y) {
                        c_index = (y * 256) + x;
                        offset = opengl_level->Textile16[i].Tile[c_index];
                        c_index = (y * 1024) + (x * 4);
                        TempTexTile[c_index + 0] = (8 * ((offset >> 10) & 0x1f));
                        TempTexTile[c_index + 1] = (8 * ((offset >> 5) & 0x1f));
                        TempTexTile[c_index + 2] = (8 * (offset & 0x1f));
                        k = TempTexTile[c_index] + TempTexTile[c_index + 1] + TempTexTile[c_index + 2];
                        if (offset & 0x8000)
                           TempTexTile[c_index + 3] = (k / 3); // set transparency based upon intensity
                        else
                           TempTexTile[c_index + 3] = 0; // ROOM_TRANSPARENCY;
                        }
                     }
                  }
               }
            }
         if (opengl_mipmap_textures) {
            gluBuild2DMipmaps(GL_TEXTURE_2D, 4,
                              256,
                              256,
                              GL_RGBA, GL_UNSIGNED_BYTE,
                              TempTexTile);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
            }
         else {
            glTexImage2D(GL_TEXTURE_2D, 0, 4,
                         256,
                         256,
                         0, GL_RGBA, GL_UNSIGNED_BYTE,
                         TempTexTile);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            }
         if (FastOpenGL) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            }
         else {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            }
         glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
         glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
         glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
         }
      TexTilesInitialised = TRUE;
      }

   /* create new list item */
   tex = (opengl_texture *)TR2alloc(sizeof(opengl_texture));

   /* get texture info */
   tex->tex = GetTextureInfo(opengl_level, type, TextureNumber, polygon_type);
   tex->w = tex->tex->Width;
   tex->h = tex->tex->Height;

   if (type != SPRITE) {
      tex->TextureID = TexTileIDs[opengl_level->ObjectTextures[TextureNumber].Tile];
      }
   else {
      tex->TextureID = TexTileIDs[opengl_level->SpriteTextures[TextureNumber].Tile];
      }

   return(tex);
}

/*
 * Initialise an OpenGL (GLubyte) structure for all sprite textures
 */
void opengl_init_sprite_textures(void)
{
   int i;

   opengl_item.SpriteTextures = TR2alloc(sizeof(opengl_texture *) * opengl_level->NumSpriteTextures);
   for (i = 0; i < (int)opengl_level->NumSpriteTextures; ++i) {
      opengl_item.SpriteTextures[i] = opengl_add_texture(SPRITE, i, RECTANGLE);
      }
}

